from daydream.plugins.datadog.nodes.base import DatadogNode


class DatadogTeam(DatadogNode):
    """A node for a Datadog team."""

    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            self.raw_data["id"],
            f"team:default/{self.raw_data['attributes']['handle']}",
        ]



import typer

DEFAULT_PROFILE = "default"
PROFILE_OPTION = typer.Option(
    ...,
    "--profile",
    envvar="DAYDREAM_PROFILE",
    help="Use profiles to manage multiple graphs",
    show_default=True,
)
PROFILE_OPTION.default = DEFAULT_PROFILE



import typer

from daydream.cli.client.logs.app import client_logs_app

client_app = typer.Typer(help="Work with clients of the Daydream MCP Server", no_args_is_help=True)
client_app.add_typer(client_logs_app, name="logs")



import rich

from daydream.cli.client.logs.app import client_logs_app


@client_logs_app.command()
def claude() -> None:
    """Show logs for Claude Desktop Daydream MCP Server"""
    rich.print("Logs are available at:")
    rich.print("~/Library/Logs/Claude/mcp.log")
    rich.print("~/Library/Logs/Claude/mcp-server-daydream.log")



from typing import Any

from daydream.knowledge import Graph, Node


class AptibleNode(Node):
    """A node representing an Aptible resource."""

    async def get_identifiers(self) -> list[str | None]:
        """Return a list of unique identifiers for the node."""
        return [
            *await super().get_identifiers(),
            self.raw_data["_links"]["self"]["href"],
        ]

    async def get_reference_identifiers(
        self,
    ) -> list[str | None | tuple[str | None, str]]:
        """Return a list of identifiers that potentially reference other nodes."""
        return [
            *await super().get_reference_identifiers(),
            *await self._get_single_resource_references(),
        ]

    async def _get_single_resource_references(self) -> list[str]:
        """Return a list of HAL links that reference a single resource."""
        single_resource_references = []
        for key, link in self.raw_data["_links"].items():
            href = link["href"]
            # If the link is a single resource (the last segment is digits), add it to the list
            if not href.strip("/").rsplit("/", 1)[0].isdigit():
                single_resource_references.append((href, f"has_{key}"))

        return single_resource_references


def inflate_resource(resource: dict[str, Any], _graph: Graph) -> AptibleNode:
    from daydream.plugins.aptible.nodes.aptible_app import AptibleApp
    from daydream.plugins.aptible.nodes.aptible_aws_instance import AptibleAwsInstance
    from daydream.plugins.aptible.nodes.aptible_container import AptibleContainer
    from daydream.plugins.aptible.nodes.aptible_custom_resource import (
        AptibleCustomResource,
    )
    from daydream.plugins.aptible.nodes.aptible_database import AptibleDatabase
    from daydream.plugins.aptible.nodes.aptible_deployment import AptibleDeployment
    from daydream.plugins.aptible.nodes.aptible_disk import AptibleDisk
    from daydream.plugins.aptible.nodes.aptible_service import AptibleService
    from daydream.plugins.aptible.nodes.aptible_vhost import AptibleVhost

    try:
        return {
            "app": AptibleApp,
            "aws_instance": AptibleAwsInstance,
            "container": AptibleContainer,
            "custom_resource": AptibleCustomResource,
            "database": AptibleDatabase,
            "deployment": AptibleDeployment,
            "disk": AptibleDisk,
            "service": AptibleService,
            "vhost": AptibleVhost,
        }[resource["_type"]](
            **{
                "node_id": str(resource["id"]),
                "raw_data": resource,
                "_graph": _graph,
            }
        )
    except KeyError as ex:
        raise LookupError(f"Unknown resource type: {resource['_type']}") from ex



from daydream.cli._app import app


@app.command()
def version() -> None:
    """
    Display the version of the Daydream CLI.
    """
    from daydream import __version__

    print(f"daydream {__version__}")



import gzip
import json
import os
from pathlib import Path
from typing import TYPE_CHECKING

from pydantic import AwareDatetime
from sh import Command, ErrorReturnCode

from daydream.plugins.aptible.logs.processing import (
    LOG_FILENAME_DATETIME_FORMAT,
    parse_log_line_timestamp,
)
from daydream.plugins.aptible.nodes.aptible_aws_instance import AptibleAwsInstance
from daydream.utils import print

if TYPE_CHECKING:
    from daydream.plugins.aptible.nodes.aptible_container import AptibleContainer


async def download_container_logs(
    stack: str,
    container: "AptibleContainer",
    start_dt: AwareDatetime,
    end_dt: AwareDatetime,
    download_dir: Path,
) -> int:
    # Stop if the container is no longer running - its logs will be in s3
    if container.raw_data["deleted_at"]:
        return 0

    # AWS Instance has the hostname of the container
    try:
        aws_instance = await anext(container.iter_neighboring(AptibleAwsInstance))
    except StopAsyncIteration:
        raise ValueError("Container has no parent AWS instance") from None

    print(
        f"Downloading container logs for {container.raw_data['docker_name']} "
        f"({aws_instance.raw_data['runtime_data']['hostname']} - {container.raw_data['layer']})"
    )

    # Download logs using pancake stack:ssh
    output_lines = await _log_lines_from_pancake(stack, container, aws_instance, start_dt, end_dt)
    if not output_lines:
        return 0

    # Write the log lines in the same format as s3 logs
    log_filename = _log_filename(container, download_dir, start_dt, end_dt)
    with gzip.open(log_filename, "wt") as f:
        for line in output_lines:
            try:
                dt, msg = line.split(" ", maxsplit=1)
                dt = parse_log_line_timestamp(dt)
                f.write(
                    json.dumps(
                        {
                            "time": dt.isoformat(),
                            "log": msg,
                        }
                    )
                    + "\n"
                )
            except ValueError:
                pass

    return 1


async def _log_lines_from_pancake(
    stack: str,
    container: "AptibleContainer",
    aws_instance: "AptibleAwsInstance",
    start_dt: AwareDatetime,
    end_dt: AwareDatetime,
) -> list[str]:
    output_lines = []
    try:
        pancake = Command(os.getenv("PANCAKE_PATH"))
        pancake_env = os.environ.copy().update(
            {
                "THOR_SILENCE_DEPRECATION": "1",
            }
        )

        output = await pancake(  # pyright: ignore [reportGeneralTypeIssues]
            "stack:ssh",
            stack,
            "--instance",
            aws_instance.raw_data["runtime_data"]["hostname"],
            "sudo",
            "docker",
            "container",
            "logs",
            container.raw_data["docker_name"],
            "--timestamps",
            "--since",
            start_dt.strftime("%Y-%m-%dT%H:%M:%SZ"),
            "--until",
            end_dt.strftime("%Y-%m-%dT%H:%M:%SZ"),
            _async=True,
            _env=pancake_env,
        )
        output_lines = output.splitlines()
    except ErrorReturnCode as e:
        print(f"Error running command: {e}")

    if len(output_lines) > 0:
        # The first line is always "Executing '...'"
        del output_lines[0]

    return output_lines


def _log_filename(
    container: "AptibleContainer",
    download_dir: Path,
    start_dt: AwareDatetime,
    end_dt: AwareDatetime,
) -> Path:
    filename_tokens = [
        f"{container.raw_data['docker_name']}-json.log",
        start_dt.strftime(LOG_FILENAME_DATETIME_FORMAT),
        end_dt.strftime(LOG_FILENAME_DATETIME_FORMAT),
        "archived.gz",
    ]

    if container.raw_data["layer"] == "proxy":
        filename_tokens.insert(0, "proxy")

    return download_dir / ".".join(filename_tokens)



import asyncio
import shlex

import anyio
import typer
from rich import print

from daydream.cli.mlflow._app import mlflow_app
from daydream.cli.options import PROFILE_OPTION
from daydream.config.utils import get_config_dir
from daydream.utils import confirm


@mlflow_app.command()
def serve(
    profile: str = PROFILE_OPTION,
    port: int = typer.Option(default=5566, help="Port for MLflow server to listen on"),
) -> None:
    async def _server() -> None:
        host = "127.0.0.1"
        config_dir = get_config_dir(profile, create=True)
        mlflow_db = config_dir / "mlflow" / "debug.db"
        mlflow_db.parent.mkdir(parents=True, exist_ok=True)
        mlflow_db.unlink(missing_ok=True)
        backend_store_uri = f"sqlite:///{mlflow_db.absolute()}"
        artifacts_dir = mlflow_db.parent / "mlartifacts"
        artifacts_dir.mkdir(parents=True, exist_ok=True)
        cmd = (
            "uvx mlflow@3 server "
            f"--host {shlex.quote(host)} --port {shlex.quote(str(port))} "
            f"--backend-store-uri {shlex.quote(backend_store_uri)} "
            f"--artifacts-destination {shlex.quote(str(artifacts_dir.absolute()))}"
        )
        print(
            f"MLflow tracking server is ready to start! Set this in your environment to use it: MLFLOW_TRACKING_URI=http://{host}:{port}"
        )
        while not await confirm("Ready to start MLflow tracking server?"):
            pass
        print(f"  ...running: {cmd}")
        proc = await asyncio.create_subprocess_shell(cmd)
        await proc.wait()

    anyio.run(_server)



import anyio

from daydream.cli.options import PROFILE_OPTION
from daydream.cli.tools._app import tools_app
from daydream.config import load_config
from daydream.config.utils import get_config_dir
from daydream.knowledge import Graph
from daydream.mcp import Context, build_mcp_server
from daydream.plugins import PluginManager


@tools_app.command("list")
def list_tools(
    profile: str = PROFILE_OPTION,
) -> None:
    """List all MCP tools available from enabled plugins."""

    async def _list_tools() -> None:
        config = load_config(profile)
        plugins = PluginManager(config=config)
        context = Context(
            profile=profile,
            config=config,
            plugins=plugins,
            graph=Graph(get_config_dir(profile) / "graph.json"),
        )
        mcp = await build_mcp_server(context)

        tools = await mcp.get_tools()
        for key, tool in tools.items():
            cmd = [key]
            for arg, arg_data in tool.parameters["properties"].items():
                arg_type = arg_data.get("type", "unknown")
                if "anyOf" in arg_data:
                    arg_type = "|".join(t["type"] for t in arg_data["anyOf"])
                cmd.append(f"<{arg}:{arg_type}>")
            print(" ".join(cmd))

    anyio.run(_list_tools)



import anyio
from rich import print

from daydream.agent.utils import get_agents
from daydream.cli.agent._app import agent_app
from daydream.cli.options import PROFILE_OPTION


@agent_app.command()
def list(profile: str = PROFILE_OPTION) -> None:
    """List the available agents."""

    async def _run() -> None:
        print("Available agents:")
        for agent in sorted(get_agents(profile)):
            print(f"* {agent}")

    anyio.run(_run)



import os

import anyio
import typer

from daydream.agent.app import listen, settings
from daydream.cli.agent._app import agent_app


@agent_app.command()
def serve(
    host: str = typer.Option(settings.DAYDREAM_HOST, help="The host to bind to"),
    port: int = typer.Option(settings.DAYDREAM_PORT, help="The port to bind to"),
    workers: int = typer.Option(settings.DAYDREAM_WORKERS, help="The number of workers to use"),
    profile: str = typer.Option(settings.DAYDREAM_PROFILE, help="The profile to use"),
    reload: bool = typer.Option(
        settings.DAYDREAM_RELOAD, help="Reload the server when the code changes"
    ),
    tunnel: bool = typer.Option(settings.DAYDREAM_TUNNEL, help="Tunnel the server through ngrok"),
    ngrok_token: str = typer.Option(
        os.getenv("NGROK_TOKEN"), help="The ngrok token to use to tunnel the server"
    ),
    ngrok_domain: str = typer.Option(
        settings.NGROK_DOMAIN, help="The ngrok domain to use to tunnel the server"
    ),
) -> None:
    async def _serve() -> None:
        await listen(
            host=host,
            port=port,
            workers=workers,
            profile=profile,
            tunnel=tunnel,
            ngrok_token=ngrok_token,
            ngrok_domain=ngrok_domain,
            reload=reload,
        )

    anyio.run(_serve)



import os
import signal

from daydream.cli.graph._app import graph_app
from daydream.cli.graph._background import cleanup_pid_file, get_pid_file, is_process_running
from daydream.cli.options import PROFILE_OPTION


@graph_app.command()
def stop(profile: str = PROFILE_OPTION) -> None:
    """Stop running graph build"""
    pid_file = get_pid_file(profile)

    if not pid_file.exists():
        print(f"No graph build running for profile '{profile}'")
        return

    try:
        pid = int(pid_file.read_text().strip())
        if is_process_running(pid):
            print(f"Stopping graph build for profile '{profile}' (PID: {pid})...")
            os.kill(pid, signal.SIGTERM)
            print("Graph build stopped successfully")
        else:
            print(f"Process not found for profile '{profile}', cleaning up stale PID file...")
        cleanup_pid_file(profile)
    except (ValueError, ProcessLookupError):
        cleanup_pid_file(profile)
        print(f"No running process found for profile '{profile}', cleaned up stale PID file")



import os
from pathlib import Path

from daydream.config.utils import get_config_dir


def get_pid_file(profile: str) -> Path:
    """Get the PID file path for a given profile graph build"""
    return get_config_dir(profile) / "graph_build.pid"


def get_log_file(profile: str) -> Path:
    """Get the log file path for a given profile graph build"""
    return get_config_dir(profile) / "graph_build.log"


def is_process_running(pid: int) -> bool:
    """Check if a process with given PID is running"""
    try:
        # Send signal 0 to check if process exists (doesn't actually kill it)
        os.kill(pid, 0)
        return True
    except (OSError, ProcessLookupError):
        return False


def check_and_create_lock(profile: str) -> bool:
    """Returns True if we can proceed, False if already running"""
    pid_file = get_pid_file(profile)

    if pid_file.exists():
        try:
            existing_pid = int(pid_file.read_text().strip())
            if is_process_running(existing_pid):
                print(f"Graph build already running for profile '{profile}' (PID: {existing_pid})")
                print(f"Use 'daydream graph stop --profile {profile}' to stop it if needed")
                return False
            else:
                # Stale PID file, remove it
                print("Found stale PID file, cleaning up...")
                pid_file.unlink()
        except (ValueError, FileNotFoundError):
            # Corrupted or missing file, remove it
            pid_file.unlink(missing_ok=True)

    return True


def create_pid_file(profile: str, pid: int) -> None:
    """Create PID file with given process ID"""
    pid_file = get_pid_file(profile)
    pid_file.parent.mkdir(parents=True, exist_ok=True)
    pid_file.write_text(str(pid))


def cleanup_pid_file(profile: str) -> None:
    get_pid_file(profile).unlink(missing_ok=True)



from datetime import UTC, datetime

from pydantic import AwareDatetime

from daydream.knowledge import HasLogs, HasMetrics
from daydream.models import LogLine, Observation
from daydream.plugins.aptible.logs.download import download_logs_for_time_range
from daydream.plugins.aptible.nodes.aptible_container import AptibleContainer
from daydream.plugins.aptible.nodes.base import AptibleNode


class AptibleService(AptibleNode, HasLogs, HasMetrics):
    """A node representing an Aptible service resource."""

    async def list_available_metrics(self) -> list[str]:
        """Return a list of available metrics for the service."""
        return ["cpu_pct", "la", "memory_all", "iops", "fs"]

    async def get_metric(
        self,
        metric_name: str,
        time_range_start: AwareDatetime,
        time_range_end: AwareDatetime,
    ) -> list[Observation] | str:
        active_containers = await self.get_active_containers(time_range_start, time_range_end)

        if not active_containers:
            return "No active containers found for the time range."

        container_observations = []
        for container in active_containers:
            observations = await container.get_metric(metric_name, time_range_start, time_range_end)

            if not isinstance(observations, list):
                continue

            container_observations.extend(observations)

        return container_observations

    async def get_logs(
        self,
        time_range_start: AwareDatetime,
        time_range_end: AwareDatetime,
    ) -> list[LogLine]:
        """Get the logs for the service."""
        active_containers = await self.get_active_containers(time_range_start, time_range_end)

        stack_name = await self._get_stack_name()
        if not stack_name:
            print(f"No stack name found for app {self.nid}. Unable to retrieve logs.")
            return []

        return await download_logs_for_time_range(
            stack=stack_name,
            node=self,
            containers=active_containers,
            start_dt=time_range_start,
            end_dt=time_range_end,
        )

    async def get_active_containers(
        self, from_datetime: AwareDatetime, to_datetime: AwareDatetime
    ) -> list[AptibleContainer]:
        """Return a list of containers associated with the service that were active during the specified time range."""
        containers = []

        async for container in self.iter_neighboring(AptibleContainer):
            active_from = datetime.fromisoformat(container.raw_data["created_at"])
            active_to = datetime.fromisoformat(
                container.raw_data.get("deleted_at") or datetime.now(UTC).isoformat()
            )

            # If the alive range overlaps the time range in any way, add the container to the list
            if (
                # Case 1: Either endpoint of the specified period falls within the active range
                (
                    active_from <= from_datetime <= active_to
                    or active_from <= to_datetime <= active_to
                )
                or
                # Case 2: The specified period falls entirely within the active range
                (from_datetime <= active_from and active_to <= to_datetime)
            ):
                containers.append(container)

        # Sort containers by their update_at timestamp, most recent first
        return sorted(containers, key=lambda c: c.raw_data["updated_at"], reverse=True)

    async def _get_stack_name(self) -> str | None:
        """Get the stack name for the service."""
        from .aptible_aws_instance import AptibleAwsInstance
        from .aptible_container import AptibleContainer

        try:
            container = await anext(self.iter_neighboring(AptibleContainer))
            aws_instance = await anext(container.iter_neighboring(AptibleAwsInstance))

            return aws_instance.raw_data["runtime_data"]["stack"]
        except (StopAsyncIteration, KeyError):
            return None



import os
import signal
from collections.abc import AsyncIterator, Awaitable, Callable
from contextlib import asynccontextmanager

import anyio
import sentry_sdk
from fastmcp import FastMCP
from mcp import ServerResult, types
from pydantic import BaseModel, ConfigDict
from pydantic_core import to_jsonable_python

from daydream.config import Config, get_config_dir, load_config
from daydream.knowledge import Graph
from daydream.plugins import PluginManager
from daydream.plugins.mixins.mcp import McpServerMixin
from daydream.telemetry import client as telemetry
from daydream.utils import print


class Context(BaseModel, extra="allow"):
    """Context for the MCP server."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    profile: str
    config: Config
    plugins: PluginManager
    mcp_server: FastMCP | None = None
    graph: Graph


async def build_mcp_server(context: Context) -> FastMCP:
    @asynccontextmanager
    async def lifespan(server: FastMCP) -> AsyncIterator[Context]:
        """Initialize the application context and plugins."""
        yield context

    mcp = FastMCP("daydream", lifespan=lifespan)

    def with_telemetry(
        handler: Callable[[types.Request], Awaitable[ServerResult]],
    ) -> Callable[[types.Request], Awaitable[ServerResult]]:
        async def _telemetry_handler(request: types.Request) -> ServerResult:
            await telemetry.send_event(
                {
                    "event": "mcp_request",
                    "request": to_jsonable_python(request),
                }
            )

            try:
                return await handler(request)
            except Exception as e:
                sentry_sdk.capture_exception(e)
                raise e

        return _telemetry_handler

    # Instrument the MCP server with Sentry
    mcp._mcp_server.request_handlers = {
        event_type: with_telemetry(handler)
        for event_type, handler in mcp._mcp_server.request_handlers.items()
    }

    for plugin in context.plugins:
        plugin.context = context

    for mcp_plugin in context.plugins.get_plugins_with_capability(McpServerMixin):
        mcp.mount(mcp_plugin.get_mcp_server(), prefix=mcp_plugin.name)

    return mcp


async def start(
    profile: str,
    disable_stdio: bool = False,
    disable_http: bool = False,
    http_host: str = "127.0.0.1",
    http_port: int = 8000,
) -> None:
    """Start the MCP server.

    This function will start the MCP server with the specified transports.

    Args:
        disable_sse: If True, the SSE transport will be disabled.
        disable_stdio: If True, the stdio transport will be disabled.
    """
    # Register direct signal handlers for immediate exit
    signal.signal(signal.SIGINT, lambda *args: os._exit(0))
    signal.signal(signal.SIGTERM, lambda *args: os._exit(0))

    config = load_config(profile)
    plugins = PluginManager(config=config)
    context = Context(
        profile=profile,
        config=config,
        plugins=plugins,
        graph=Graph(get_config_dir(profile) / "graph.json"),
    )

    mcp = await build_mcp_server(context)

    async def _run_stdio_server() -> None:
        await mcp.run_stdio_async(show_banner=False)

    async def _run_http_server() -> None:
        await mcp.run_http_async(
            show_banner=False,
            transport="http",
            host=http_host,
            port=http_port,
        )

    async with anyio.create_task_group() as tg:
        if not disable_stdio:
            tg.start_soon(_run_stdio_server)
        if not disable_http:
            tg.start_soon(_run_http_server)

    print("MCP server stopped")



from .utils import (
    CONFIG_ROOT,
    DEFAULT_CONFIG_FILE,
    Config,
    PluginConfig,
    get_config_dir,
    load_config,
)

__all__ = (
    "CONFIG_ROOT",
    "DEFAULT_CONFIG_FILE",
    "Config",
    "PluginConfig",
    "get_config_dir",
    "load_config",
)



import anyio

from daydream.agent.utils import delete_agent
from daydream.cli.agent._app import agent_app
from daydream.cli.options import PROFILE_OPTION


@agent_app.command()
def delete(agent_name: str, profile: str = PROFILE_OPTION) -> None:
    """Delete an agent."""

    async def _run() -> None:
        delete_agent(agent_name, profile)

    anyio.run(_run)



from .graph import KnowledgeGraphMixin
from .mcp import McpServerMixin, prompt, resource, tool

__all__ = [
    "KnowledgeGraphMixin",
    "McpServerMixin",
    "prompt",
    "resource",
    "tool",
]



from importlib.metadata import version as get_version
from typing import cast

from dotenv import load_dotenv

from .sentry import *  # noqa: F403
from .warnings import *  # noqa: F403

__version__ = get_version(cast("str", __package__))

load_dotenv()



from typing import TYPE_CHECKING, cast
from fastmcp import Client
import pytest
from freezegun.api import FrozenDateTimeFactory

if TYPE_CHECKING:
    from mcp.types import TextContent


@pytest.mark.asyncio
async def test_core_calculate(mcp_client: Client) -> None:
    result = await mcp_client.call_tool("core_calculate", {"expression": "1 + 1"})
    result = cast("TextContent", result.content[0])
    assert float(result.text) == 2.0


@pytest.mark.asyncio
async def test_core_current_datetime(mcp_client: Client, freezer: FrozenDateTimeFactory) -> None:
    current_datetime = "2025-01-01T12:00:00Z"
    freezer.move_to(current_datetime)
    result = await mcp_client.call_tool("core_current_datetime")
    result = cast("TextContent", result.content[0])
    assert result.text.strip('"') == current_datetime



from datetime import datetime

import msgspec
from pydantic import AwareDatetime, BaseModel


class LogAnomaly(BaseModel):
    pattern: str
    peak_increase: float
    peak_time: AwareDatetime
    counts_sparkline: str


class LogLine(msgspec.Struct):
    time: datetime
    log: str


class Observation(BaseModel):
    node_id: str
    observation_type: str
    data: dict[AwareDatetime, float]



from .edges import Edge
from .graph import Graph
from .nodes import NODE_REGISTRY, Node
from .nodes.mixins import HasLogs, HasMetrics

__all__ = [
    "NODE_REGISTRY",
    "Edge",
    "Graph",
    "HasLogs",
    "HasMetrics",
    "Node",
]



from typing import Any

import questionary

from daydream.config.utils import Config, PluginSettings
from daydream.plugins.base import Plugin
from daydream.utils import classproperty


class LlmPlugin(Plugin):
    """A plugin for configuring LLM models."""

    def __init__(
        self,
        *args: Any,
        model: str,
        api_key: str,
        temperature: float,
        max_tokens: int,
        cache: bool,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.model = model
        self.api_key = api_key
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.cache = cache

    @classproperty
    def default_plugin_settings(cls) -> PluginSettings:
        return {
            "model": "openai/gpt-4o-mini",
            "api_key": "",
            "temperature": 0,
            "max_tokens": 8192,
            "cache": True,
        }

    async def interactive_configure(self, cfg: Config) -> PluginSettings:
        """Interactive wizard for configuring the settings of this plugin."""
        defaults = self.default_plugin_settings

        return {
            "model": await questionary.text("Model", default=defaults["model"]).unsafe_ask_async(),
            "api_key": await questionary.password(
                "API key",
                default=defaults["api_key"],
            ).unsafe_ask_async(),
            "temperature": float(
                await questionary.text(
                    "Model temperature (any float between 0 and 2)",
                    default=str(defaults["temperature"]),
                    validate=lambda v: v.replace(".", "").isdigit() and 0 <= float(v) <= 2,
                ).unsafe_ask_async()
            ),
            "max_tokens": int(
                await questionary.text(
                    "Maximum number of tokens to generate for each response",
                    default=str(defaults["max_tokens"]),
                    validate=lambda v: v.isdigit() and 0 < int(v) <= 8192,
                ).unsafe_ask_async()
            ),
            "cache": await questionary.confirm(
                "Enable caching of responses",
                default=defaults["cache"],
            ).unsafe_ask_async(),
        }



from pydantic import AwareDatetime

from daydream.knowledge import HasLogs, HasMetrics
from daydream.models import LogLine, Observation
from daydream.plugins.aptible.client import AptibleMetricsClient
from daydream.plugins.aptible.logs.download import download_logs_for_time_range
from daydream.plugins.aptible.nodes.base import AptibleNode


class AptibleContainer(AptibleNode, HasMetrics, HasLogs):
    """A container running on an Aptible service."""

    async def get_identifiers(self) -> list[str | None]:
        """Return a list of unique identifiers for the container."""
        return [
            *await super().get_identifiers(),
            self.raw_data["docker_name"],
        ]

    async def get_reference_identifiers(
        self,
    ) -> list[str | None | tuple[str | None, str]]:
        """Return a list of identifiers that potentially reference other nodes."""
        return [
            *await super().get_reference_identifiers(),
            (self.raw_data["aws_instance_id"], "running_on"),
            (self.raw_data["host"], "running_on"),
        ]

    async def list_available_metrics(self) -> list[str]:
        """Return a list of available metrics for the container."""
        metrics = ["cpu_pct", "la", "memory_all"]

        # If the container has mounts, add disk-related metrics.
        if self.raw_data["mounts"]:
            metrics = [*metrics, "iops", "fs"]

        return metrics

    async def get_metric(
        self,
        metric_name: str,
        time_range_start: AwareDatetime,
        time_range_end: AwareDatetime,
    ) -> list[Observation] | str:
        """Return a metric for the container."""
        metrics_client = AptibleMetricsClient()

        container_id = str(self.raw_data["id"])
        observations = await metrics_client.get_metrics(
            self.nid, [container_id], metric_name, time_range_start, time_range_end
        )
        return observations or "No metrics data found for the time range."

    async def get_logs(
        self,
        time_range_start: AwareDatetime,
        time_range_end: AwareDatetime,
    ) -> list[LogLine]:
        """Get the logs for the container."""
        stack_name = await self._get_stack_name()
        if stack_name is None:
            return []

        return await download_logs_for_time_range(
            stack=stack_name,
            node=self,
            containers=[self],
            start_dt=time_range_start,
            end_dt=time_range_end,
        )

    async def _get_stack_name(self) -> str | None:
        """Get the stack name for the container."""
        from .aptible_aws_instance import AptibleAwsInstance

        try:
            aws_instance = await anext(self.iter_neighboring(AptibleAwsInstance))
            return aws_instance.raw_data["runtime_data"]["stack"]
        except (StopAsyncIteration, KeyError):
            return None



from daydream.knowledge.graph import Graph
from daydream.plugins.kubernetes.nodes.kubernetes_node import KubernetesNode
from daydream.plugins.kubernetes.nodes.utils import label_key_value_to_node_id

from .base import KubernetesBaseNode


class KubernetesPod(KubernetesBaseNode):
    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            *[i.get("ip") for i in self.raw_data.get("status", {}).get("podIPs", []) if "ip" in i],
            *[
                label_key_value_to_node_id(key, value)
                for key, value in self.raw_data.get("metadata", {}).get("labels", {}).items()
            ],
        ]

    async def get_reference_identifiers(self) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
            (self.raw_data.get("spec", {}).get("nodeName"), "is_on_node"),
            (self.raw_data.get("spec", {}).get("serviceAccount"), "has_service_account"),
            (
                KubernetesNode(
                    node_id=self.raw_data.get("spec", {}).get("nodeName"), _graph=Graph()
                ).nid,
                "running_on",
            ),
        ]



import os

import questionary
from pydantic import AwareDatetime, BaseModel

from daydream.config.utils import Config, PluginSettings
from daydream.plugins.base import Plugin
from daydream.plugins.mixins import McpServerMixin, tool
from daydream.plugins.papertrail.client import PapertrailClient, PapertrailLogEvent

CLAUDE_DESKTOP_RESULT_MAX_LENGTH = 1048576
RESULT_LIMIT = int(CLAUDE_DESKTOP_RESULT_MAX_LENGTH * 0.9)


class PapertrailSearchResult(BaseModel):
    """Result of a Papertrail log search."""

    """Number of logs that were found, before any truncation"""
    count: int
    """True when the search results were truncated due to the response size limit (count is not changed when this happens)"""
    truncated: bool = False
    """The log events that were found"""
    results: list[PapertrailLogEvent]


class PapertrailPlugin(Plugin, McpServerMixin):
    _client: PapertrailClient

    def init_plugin(self) -> None:
        super().init_plugin()
        self._client = PapertrailClient(
            token=self._settings.get("token")
            or os.environ.get("PAPERTRAIL_TOKEN", "")
            or os.environ.get("PAPERTRAIL_API_TOKEN", ""),
        )

    async def validate_plugin_config(self) -> None:
        await super().validate_plugin_config()
        await self._client.verify_connection()

    async def interactive_configure(self, cfg: Config) -> PluginSettings:
        """Interactive wizard for configuring the settings of this plugin."""
        existing_token = os.environ.get("PAPERTRAIL_TOKEN")
        if existing_token is not None:
            return {
                "token": await questionary.text(
                    "Enter your Papertrail API token (default from PAPERTRAIL_TOKEN env var)",
                    default=existing_token,
                ).unsafe_ask_async(),
            }
        return {
            "token": await questionary.text(
                "Enter your Papertrail API token",
            ).unsafe_ask_async(),
        }

    @tool()
    async def search_logs(
        self,
        query: str,
        min_time: AwareDatetime,
        max_time: AwareDatetime,
    ) -> PapertrailSearchResult:
        """Search Papertrail for logs within a given time range

        Args:
            query (str): The search query.
            min_time (AwareDatetime): The starting time for the range within which to search.
            max_time (AwareDatetime): The ending time for the range within which to search.

        Returns:
            PapertrailSearchResult: logs that matched the query and fit within response limit
        """
        logs = sorted(
            [
                log
                async for log in self._client.search(
                    query=query,
                    min_time=min_time,
                    max_time=max_time,
                )
            ],
            key=lambda x: x.received_at,
        )
        result = PapertrailSearchResult(results=logs, count=len(logs))
        content_length = len(result.model_dump_json(indent=2))
        if content_length <= RESULT_LIMIT:
            return result
        result.truncated = True
        while result.results and content_length > RESULT_LIMIT:
            result.results.pop()
            content_length = len(result.model_dump_json(indent=2))
        return result



from daydream.plugins.kubernetes.nodes.utils import label_key_value_to_node_id

from .base import KubernetesBaseNode


class KubernetesNode(KubernetesBaseNode):
    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            *[
                label_key_value_to_node_id(key, value)
                for key, value in self.raw_data.get("metadata", {}).get("labels", {}).items()
            ],
        ]

    async def get_reference_identifiers(self) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
            *[
                (a["address"], "running_on")
                for a in self.raw_data.get("status", {}).get("addresses", [])
                if "address" in a
            ],
        ]



import shutil
import subprocess

import typer

from daydream.cli.graph._app import graph_app
from daydream.cli.graph._background import get_log_file
from daydream.cli.options import PROFILE_OPTION


@graph_app.command()
def logs(
    profile: str = PROFILE_OPTION,
    follow: bool = typer.Option(False, "--follow", "-f", help="Follow log output"),
) -> None:
    """View graph build logs"""
    log_file = get_log_file(profile)
    tail_cmd = shutil.which("tail")
    if not tail_cmd:
        print("'tail' command not found. Please install it.")
        return

    if not log_file.exists():
        print(f"No log file found for profile '{profile}'")
        print(f"Expected location: {log_file}")
        return

    if follow:
        print(f"Following logs for profile '{profile}' (Ctrl+C to stop)")
        print(f"Log file: {log_file}")

        try:
            subprocess.run([tail_cmd, "-f", str(log_file)])
        except KeyboardInterrupt:
            print("\nStopped following logs")
    else:
        print(f"Recent logs for profile '{profile}':")
        print(f"Log file: {log_file}")

        result = subprocess.run(
            [tail_cmd, "-n", "50", str(log_file)], capture_output=True, text=True
        )
        print(result.stdout)



import sys

import anyio
import typer
from rich import print

from daydream.agent.analysis import AnalysisAgent
from daydream.agent.utils import load_agent
from daydream.cli.agent._app import agent_app
from daydream.cli.options import PROFILE_OPTION


@agent_app.command(
    help="Run an agent with the provided payload and print the analysis. A payload can be passed as an argument or piped to stdin."
)
def run(
    agent_name: str = typer.Argument(..., help="The name of the agent to run"),
    payload: str | None = typer.Argument(
        None,
        help="The alert payload to analyze. Alternatively, you can pipe the payload to stdin.",
    ),
    profile: str = PROFILE_OPTION,
    debug: bool = typer.Option(
        False,
        help="Enable debug mode to print the history of the agent.",
    ),
) -> None:
    """Run an agent with the provided payload and print the analysis."""

    async def _run() -> None:
        # Read data from stdin if it's being piped to us.
        if not sys.stdin.isatty():
            if payload is not None:
                print("[red]Cannot pass a payload argument when piping data to stdin.[/red]")
                raise typer.Exit(code=1)
            data = sys.stdin.read().strip()
        else:
            # Otherwise, use the payload argument.
            data = payload

        if not data:
            print("[red]No payload provided.[/red]")
            print(
                "[bold]Pass an alert payload as an argument or pipe the payload data to stdin.[/bold]"
            )
            raise typer.Exit(code=1)

        # Get the config directory and load the specific agent
        try:
            agent = load_agent(agent_name, profile)
        except FileNotFoundError as ex:
            print(f"[red]Agent {agent_name!r} not found at {str(ex.filename)!r}[/red]")
            print(f"[bold]Use 'daydream agent create {agent_name!r}' to create a new agent.[/bold]")
            raise typer.Exit(code=1) from ex

        # Run the analysis with the specific agent
        analysis_agent = AnalysisAgent(profile)
        try:
            result = await analysis_agent.acall(payload=data, agent=agent)
            print(result)
        except Exception as ex:
            print(f"[red]Analysis failed:[/red] {ex}")
            raise typer.Exit(code=1) from ex
        finally:
            if debug:
                print("\n\n===== DEBUG OUTPUT =====\n")
                analysis_agent.inspect_history(n=1000)
                print("\n===== END DEBUG OUTPUT =====\n\n")

    anyio.run(_run)



import asyncio

import sentry_sdk
from sentry_sdk.integrations.typer import TyperIntegration


async def init() -> None:
    sentry_sdk.init(
        dsn="https://355c313b713086f073fc60c6527f9db8@o12041.ingest.us.sentry.io/4509431193272320",
        # Add data like request headers and IP for users,
        # see https://docs.sentry.io/platforms/python/data-management/data-collected/ for more info
        send_default_pii=True,
        integrations=[TyperIntegration()],
    )


# Only run if not in an existing event loop
if not asyncio.get_event_loop_policy():
    asyncio.run(init())



from typing import cast

from pydantic import AnyUrl, ValidationError

from .base import AptibleNode


class AptibleApp(AptibleNode):
    """A node representing an Aptible app resource."""

    async def get_identifiers(self) -> list[str | None]:
        """Return a list of unique identifiers for the node."""
        return [
            *await super().get_identifiers(),
            self.raw_data["handle"],
        ]

    async def get_reference_identifiers(
        self,
    ) -> list[str | None | tuple[str | None, str]]:
        """Return a list of identifiers that potentially reference other nodes."""
        return [
            *await super().get_reference_identifiers(),
            *await self._get_hostname_references(),
        ]

    async def _get_hostname_references(self) -> set[tuple[str, str]]:
        """Return a set of hostnames referenced by the app."""
        hostnames = set()

        try:
            env = cast("dict[str, str]", self.raw_data["_embedded"]["current_configuration"]["env"])
        except KeyError:
            return hostnames

        for value in env.values():
            try:
                url = AnyUrl(value)
            except ValidationError:
                continue
            if not url.host:
                continue
            hostnames.add(url.host)

        return {(hostname, "depends_on") for hostname in hostnames}



import asyncio
import os

import typer
from rich import print

from daydream.agent.utils import get_agent_template
from daydream.cli.agent._app import agent_app
from daydream.cli.options import PROFILE_OPTION
from daydream.config.utils import get_config_dir
from daydream.utils import edit_file


@agent_app.command()
def edit(
    agent_name: str = typer.Argument(..., help="The name of the agent to edit"),
    profile: str = PROFILE_OPTION,
    editor: str = typer.Option(
        os.environ.get("EDITOR"), help="The editor to use to open the agent file"
    ),
) -> None:
    """Edit an existing agent configuration file."""

    async def _edit() -> None:
        # Get the config directory for the profile
        config_dir = get_config_dir(profile, create=False)

        # Build the agent file path
        agent_file = config_dir / "agents" / f"{agent_name}.yaml"

        # If they're editing the default agent and it doesn't exist, create it.
        if agent_name == "default" and not agent_file.exists():
            agent_file = config_dir / "agents" / "default.yaml"
            agent_file.parent.mkdir(parents=True, exist_ok=True)
            agent_file.touch()
            agent_file.write_text(get_agent_template(agent_name))

        # Check if the agent file exists
        if not agent_file.exists():
            print(f"Agent '{agent_name}' not found at {agent_file}")
            print(f"Use 'daydream agent create {agent_name}' to create a new agent.")
            raise typer.Abort()

        # Open the file in the user's editor
        try:
            await edit_file(agent_file, editor)
        except ValueError as ex:
            print(
                "[red]No editor specified. Set the $EDITOR environment variable or use --editor option.[/red]"
            )
            print(f"[blue]Please manually open {str(agent_file)!r} in your editor.[/blue]")
            raise typer.Exit() from ex

    asyncio.run(_edit())



from collections import defaultdict

import anyio

from daydream.knowledge import Graph, Node
from daydream.knowledge.edges import Edge
from daydream.plugins import Plugin
from daydream.plugins.aptible.client import AptibleClient
from daydream.plugins.aptible.nodes.base import inflate_resource
from daydream.plugins.mixins import KnowledgeGraphMixin
from daydream.utils import print


class AptiblePlugin(Plugin, KnowledgeGraphMixin):
    """A plugin for the Aptible PaaS."""

    async def validate_plugin_config(self) -> None:
        await super().validate_plugin_config()
        async with AptibleClient() as client:
            await client.validate_auth()

    async def populate_graph(self, graph: Graph) -> None:
        """Initialize the graph with nodes and edges from the Aptible API."""
        self.client = AptibleClient()

        edge_count = 0
        seen_nodes = defaultdict(set)

        async with self.client.concurrent_paginator("/edges") as (
            edges,
            total_count,
        ):
            notify_threshold = max(int(total_count * 0.01), 1)

            async for edge in edges:
                source_node = inflate_resource(
                    edge["_embedded"]["source_resource"],
                    _graph=graph,
                )
                destination_node = inflate_resource(
                    edge["_embedded"]["destination_resource"],
                    _graph=graph,
                )

                await graph.add_edge(
                    Edge(
                        source_node=source_node,
                        destination_node=destination_node,
                        properties={
                            "relationship_type": edge["relationship_type"],
                        },
                    )
                )

                if source_node.nid not in seen_nodes[source_node.node_type]:
                    seen_nodes[source_node.node_type].add(source_node.nid)
                if destination_node.nid not in seen_nodes[destination_node.node_type]:
                    seen_nodes[destination_node.node_type].add(destination_node.nid)

                edge_count += 1

                if edge_count % notify_threshold == 0:
                    percent_complete = (edge_count / total_count) * 100
                    print(f"Edge loading: {edge_count}/{total_count} ({percent_complete:.2f}%)")

        # Update app configurations
        async def _update_app_configuration(node: Node) -> None:
            # If the app has no link to a current_configuration, skip it.
            if "current_configuration" not in node.raw_data["_links"]:
                return

            # If the app has no services, it wont have a configuration, so skip it.
            if not node.raw_data["_embedded"]["services"]:
                return

            print(f"Retrieving the current configuration for {node.nid}")
            response = await self.client.get(
                node.raw_data["_links"]["current_configuration"]["href"]
            )

            if response.status_code != 200:
                print(f"Failed to retrieve the current configuration for {node.nid}")
                return

            node.raw_data["_embedded"]["current_configuration"] = response.json()

        async with anyio.create_task_group() as tg:
            async for node in graph.iter_nodes():
                if node.node_type == "aptible_app":
                    tg.start_soon(_update_app_configuration, node)

        total_nodes = sum(len(seen_nodes[node_type]) for node_type in seen_nodes)

        print(f"Initialized {edge_count} edges between {total_nodes} nodes")
        for node_type, node_ids in seen_nodes.items():
            print(f"  - {node_type}: {len(node_ids)}")



import json
from typing import Annotated

import anyio
import typer
from fastmcp import Client
from mcp.types import TextContent

from daydream.cli.options import PROFILE_OPTION
from daydream.cli.tools._app import tools_app
from daydream.config import load_config
from daydream.config.utils import get_config_dir
from daydream.knowledge import Graph
from daydream.mcp import Context, build_mcp_server
from daydream.plugins import PluginManager


@tools_app.command()
def call(
    tool: str,
    count_results: Annotated[
        bool,
        typer.Option(
            "--count-results",
            "-c",
            help="Count the number of results returned by the tool instead of printing the results (works best with list or dict tool responses)",
        ),
    ] = False,
    count_level: Annotated[
        int, typer.Option("--count-level", help="Count level for nested structures", min=0, max=1)
    ] = 0,
    arguments: Annotated[list[str] | None, typer.Argument()] = None,
    profile: str = PROFILE_OPTION,
) -> None:
    """Call an MCP tool from the command line."""

    async def _call_tool() -> None:
        config = load_config(profile)
        plugins = PluginManager(config=config)
        context = Context(
            profile=profile,
            config=config,
            plugins=plugins,
            graph=Graph(get_config_dir(profile) / "graph.json"),
        )
        mcp = await build_mcp_server(context)

        tools = await mcp.get_tools()
        if tool not in tools:
            raise typer.Abort(f"Tool {tool} not found")

        tool_def = tools[tool]
        tool_args = {}
        if arguments:
            # Convert a list of arguments into a dict
            for k, v in zip(tool_def.parameters["properties"].keys(), arguments, strict=True):
                # Try decoding as JSON to handle complex arguments (lists, dicts, etc.)
                try:
                    tool_args[k] = json.loads(v)
                except json.JSONDecodeError:
                    tool_args[k] = v

        client = Client(mcp)
        async with client:
            result = await client.call_tool(tool, tool_args)
            try:
                content = next(r.text for r in result.content if isinstance(r, TextContent))
                if count_results and count_level == 0:
                    print(
                        json.dumps(
                            {"count": len(json.loads(content)), "content_length": len(content)},
                            indent=2,
                        )
                    )
                elif count_results and count_level == 1:
                    print(
                        json.dumps(
                            {
                                k: (
                                    {
                                        "count": len(v),
                                        "content_length": len(json.dumps(v)),
                                    }
                                    if isinstance(v, list | dict)
                                    else v
                                )
                                for k, v in json.loads(content).items()
                            },
                            indent=2,
                        )
                    )
                else:
                    print(content)
            except StopIteration:
                print(f"{tool} returned no printable results")

    anyio.run(_call_tool)



from daydream.cli.graph._app import graph_app
from daydream.cli.graph._background import (
    cleanup_pid_file,
    get_log_file,
    get_pid_file,
    is_process_running,
)
from daydream.cli.options import PROFILE_OPTION


@graph_app.command()
def status(profile: str = PROFILE_OPTION) -> None:
    """Check if graph build is running"""
    pid_file = get_pid_file(profile)

    if not pid_file.exists():
        print(f"No graph build running for profile '{profile}'")
        return

    try:
        pid = int(pid_file.read_text().strip())
        if is_process_running(pid):
            print(f"Graph build running for profile '{profile}' (PID: {pid})")

            # Show log file info if it exists
            log_file = get_log_file(profile)
            if log_file.exists():
                print(f"View logs: daydream graph logs --profile {profile} --follow")
        else:
            print(f"Stale PID file found for profile '{profile}', cleaning up...")
            cleanup_pid_file(profile)
    except ValueError:
        print(f"Corrupted PID file found for profile '{profile}', cleaning up...")
        cleanup_pid_file(profile)
